class Event(list):
    def __call__(self, *args, **kwargs):
        for f in self:
            f(*args, **kwargs)

    def __repr__(self):
        return "Event(%s)" % list.__repr__(self)


class MDownloadEventArgs:
    def __init__(self):
        self.filekind = ""
        self.filename = ""
        self.maxvalue = 1
        self.currentvalue = 1




from pmlauncher import pml, mlogin, mlaunchoption
import subprocess
import os


# initialize
# p = os.environ["appdata"] + "\\.minecraft"  # windows default game directory
# p = os.getcwd() + "/game"
p = os.path.abspath("/home/myu/.minecraft")
pml.initialize(p)
print("Initialized in " + pml.getGamePath())


# login
print("session : test user (tester123)")
session = mlogin.session()
session.username = "tester123"
session.uuid = "uuid"
session.access_token = "access_token"


# get profiles
profiles = pml.updateProfiles()
for item in profiles:
    print(item.name)

inputVersion = input("input version : ")


# download event handler
# filekind : library , minecraft, index, resource
def downloadEvent(x):
    print(x.filekind + " - " + x.filename + " - " + str(x.currentvalue) + "/" + str(x.maxvalue))


pml.downloadEventHandler = downloadEvent

# download profile and create argument
args = pml.startProfile(inputVersion, 
                        xmx_mb=1024,
                        session=session,

                        launcher_name="pml",  # option
                        server_ip="",
                        jvm_args="",
                        screen_width=0,
                        screen_height=0)


# start process
with open("args.txt", "w") as f:  # for debug
    f.write(args)
print(args)

# in linux system, use os.system instead of subprocess
os.system("java " + args)
#mc = subprocess.Popen("java " + args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=pml.getGamePath())

print("launched!")


# write output
with mc.stdout as gameLog:
    while True:
        line = gameLog.readline()
        if not line:
            break
        print(line)

if mc.returncode:
    print(f"Client returned {mc.returncode}!")




import platform
import os

checkOSRules = True

is64bit = platform.machine().endswith('64')
if is64bit:
    arch = "64"
else:
    arch = "32"

osversion = platform.release()
osname = ""
sysname = platform.system()
if sysname == "Linux":
    osname = "linux"
elif sysname == "Darwin":
    osname = "osx"
else:
    osname = "windows"
#osname = "windows"  # fake os to debug


def checkAllowOS(arr):
    for job in arr:
        action = True  # allow / disallow
        containCurrentOS = True

        for key, value in job.items():

            if key == "action":
                if value == "allow":
                    action = True
                else:
                    action = False

            elif key == "os":
                containCurrentOS = check_os_contains(value.items())

            elif key == "features":
                return False

        if not action and containCurrentOS:
            return False
        elif action and containCurrentOS:
            return True
        elif action and not containCurrentOS:
            return False


def check_os_contains(arr):
    for osKey, osValue in arr:
        if osKey == "name" and osValue == osname:
            return True
    return False



from . import *

profiles = None
downloadEventHandler = None


def getGamePath():
    return minecraft.path


def initialize(path):
    minecraft.initialize(path)


def updateProfiles():
    global profiles
    profiles = mprofileinfo.getProfiles()
    return profiles


def getProfile(name):
    global profiles
    if profiles is None:
        updateProfiles()

    return mprofile.get_profile(profiles, name)


def downloadProfile(profile, downloadAssets = True):
    if type(profile) is not mprofile.profile:
        raise ValueError("profile must be mprofile.profile type")

    downloader = mdownloader.mdownload(profile)
    if downloadEventHandler:
        downloader.downloadFileChangedEvent.append(downloadEventHandler)
    downloader.downloadAll(downloadAssets)


def startProfile(name, **option):
    profile = getProfile(name)
    print(profile.jvm_arguments)
    downloadProfile(profile)

    l = mlaunchoption.launchoption()
    if option.get("launchoption"):
        l = option.get("launchoption")
    else:
        l.__dict__.update(option)
    
    l.start_profile = profile
    launch = mlaunch.launch(l)
    return launch.createProcess()



from pmlauncher import minecraft, mrule
import os

class mlibrary:
    def __init__(self):
        self.isNative = False
        self.name = ""
        self.path = ""
        self.url = ""
        self.isRequire = True
        self.hash = ""


checkOSRules = True
defaultLibraryServer = "https://libraries.minecraft.net/"

def nameToPath(name, native):  # library name to relative path
    try:
        tmp = name.split(':')
        front = tmp[0].replace('.', '/')
        back = ""

        for i in range(1, len(tmp)):
            if i == len(tmp) - 1:
                back += tmp[i]
            else:
                back += tmp[i] + ":"

        libpath = front + "/" + back.replace(":", "/") + "/" + (back.replace(":", "-"))
        if native:
            libpath += "-" + native + ".jar"
        else:
            libpath += ".jar"
        return libpath
    except Exception:
        return ""


def createLibrary(name, nativeId, job):
    path = job.get("path")
    if not path:
        path = nameToPath(name, nativeId)

    url = job.get("url")
    if not url:
        url = defaultLibraryServer + path
    elif not url.split('/')[-1]:
        url += path

    library = mlibrary()
    library.hash = job.get("sha1")
    library.name = name
    library.path = os.path.normpath(minecraft.library + "/" + path)
    library.url = url
    if nativeId:
        library.isNative = True
    else:
        library.isNative = False

    return library

def parselist(json):
    list = []

    for item in json:
            name = item.get("name")
            if name is None:
                continue

            # check rules
            rules = item.get("rules")
            if checkOSRules and rules:
                isRequire = mrule.checkAllowOS(rules)

                if not isRequire:
                    continue

            # forge library
            downloads = item.get("downloads")
            if not downloads:  # downloads == null
                natives = item.get("natives")

                nativeId = None
                if natives is not None:  # natives != null
                    nativeId = natives.get(mrule.osname)

                list.append(createLibrary(name, nativeId, item))
                continue

            # native library
            classif = downloads.get("classifiers")
            if classif:
                native_id = None
                native_obj = item.get("natives")
                if native_obj:
                    native_id = native_obj.get(mrule.osname).replace("${arch}", mrule.arch)

                if native_id and classif.get(native_id):
                    job = classif.get(native_id)
                    list.append(createLibrary(name, native_id, job))

            # common library
            arti = downloads.get("artifact")
            if arti:
                list.append(createLibrary(name, "", arti))

    return list




class launchoption:
    def __init__(self):
        self.xmx_mb = 1024
        self.start_profile = None
        self.session = None
        self.launcher_name = ""
        self.server_ip = ""
        self.jvm_arg = ""
        self.screen_width = 0
        self.screen_height = 0

    def checkValid(self):
        exMsg = ""
        if not self.xmx_mb:
            exMsg = "xmx_mb is too small"
        if not self.start_profile:
            exMsg = "start_profile was None"
        if not self.session:
            exMsg = "session was None"
        if " " in self.launcher_name:
            exMsg = "launcher_name cannot contain space character"

        if exMsg:
            raise ValueError(exMsg)




import os
from pmlauncher import minecraft
import json
import requests


class mprofileinfo:
    def __init__(self):
        self.isweb = True
        self.name = ""
        self.path = ""


def getProfilesFromLocal():
    files = os.listdir(minecraft.version)
    arr = list()

    if not files:
        return arr

    for item in files:
        filepath = os.path.normpath(minecraft.version + "/" + item + "/" + item + ".json")
        if os.path.isfile(filepath):
            profile = mprofileinfo()
            profile.isweb = False
            profile.name = item
            profile.path = filepath
            arr.append(profile)

    return arr


def getProfilesFromWeb():
    result = list()
    url = "https://launchermeta.mojang.com/mc/game/version_manifest.json"
    jarr = json.loads(requests.get(url).text)
    for item in jarr.get("versions"):
        profile = mprofileinfo()
        profile.isweb = True
        profile.name = item.get("id")
        profile.path = item.get("url")
        result.append(profile)

    return result


def getProfiles():
    arr = getProfilesFromLocal()
    for item1 in getProfilesFromWeb():
        exist = False
        for item2 in arr:
            if item1.name == item2.name:
                exist = True
                break
        if not exist:
            arr.append(item1)
    return arr


import os

path = ""
library = ""
version = ""
assets = ""
index = ""
assetObject = ""
assetLegacy = ""
natives = ""


def initialize(_path):
    global path, library, version, assets, index, assetObject, assetLegacy, natives

    path = os.path.normpath(_path)
    library = os.path.normpath(path + "/libraries")
    version = os.path.normpath(path + "/versions")
    assets = os.path.normpath(path + "/assets")
    index = os.path.normpath(assets + "/indexes")
    assetObject = os.path.normpath(assets + "/objects")
    assetLegacy = os.path.normpath(assets + "/virtual/legacy")
    natives = os.path.normpath(path + "/natives")

    mkd(path)
    mkd(library)
    mkd(version)
    mkd(index)
    mkd(assetLegacy)
    mkd(assetObject)
    mkd(natives)


def mkd(p):
    if not os.path.isdir(p):
        os.makedirs(p)



import os
import requests
from pmlauncher import minecraft, mevent
import json
from shutil import copyfile
import hashlib


def mkd(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def download(url, path):
    dirpath = os.path.dirname(path)
    mkd(dirpath)

    response = requests.get(url)
    if int(response.status_code / 100) is not 2:
        return

    f = open(path, "wb")
    f.write(response.content)
    f.close()


class mdownload:
    def __init__(self, _profile):
        self.checkHash = True
        self.profile = _profile
        self.doFireEvents = True
        self.downloadFileChangedEvent = mevent.Event()

    def fireEvent(self, kind, name, max, current):
        if not self.doFireEvents:
            return

        args = mevent.MDownloadEventArgs()
        args.filekind = kind
        args.filename = name
        args.maxvalue = max
        args.currentvalue = current
        self.downloadFileChangedEvent(args)

    def checkFileSHA1(self, path, fhash):
        if not self.checkHash:
            return True
        if not fhash:
            return True

        f = open(path, "rb")
        data = f.read()
        f.close()

        return fhash == hashlib.sha1(data).hexdigest()

    def checkFileValidation(self, path, fhash):
        return os.path.isfile(path) and self.checkFileSHA1(path, fhash)

    def downloadAll(self, downloadAssets):
        self.downloadLibraries()
        self.downloadMinecraft()
        if downloadAssets:
            self.downloadIndex()
            self.downloadResources()

    def downloadLibraries(self):
        count = len(self.profile.libraries)
        for i in range(0, count):
            lib = self.profile.libraries[i]
            if lib.isRequire and lib.path and lib.url and not self.checkFileValidation(lib.path, lib.hash):
                download(lib.url, lib.path)

            self.fireEvent("library", lib.name, count, i + 1)

    def downloadIndex(self):
        path = os.path.normpath(minecraft.index + "/" + self.profile.assetId + ".json")
        if self.profile.assetUrl and not self.checkFileValidation(path, self.profile.assetHash):
            download(self.profile.assetUrl, path)

        self.fireEvent("index", self.profile.assetId, 1, 1)

    def downloadResources(self):
        indexPath = os.path.normpath(minecraft.index + "/" + self.profile.assetId + ".json")
        if not os.path.isfile(indexPath):
            return

        f = open(indexPath, "r")
        content = f.read()
        f.close()

        index = json.loads(content)

        isVirtual = False
        v = index.get("virtual")
        if v and v == "true":
            isVirtual = True

        items = list(index.get("objects").items())
        count = len(items)
        for i in range(0, count):
            key = items[i][0]
            value = items[i][1]

            hash = value.get("hash")
            hashName = hash[:2] + "/" + hash
            hashPath = os.path.normpath(minecraft.assetObject + "/" + hashName)
            hashUrl = "http://resources.download.minecraft.net/" + hashName

            if not os.path.isfile(hashPath):
                download(hashUrl, hashPath)

            if isVirtual:
                resPath = os.path.normpath(minecraft.assetLegacy + "/" + key)

                if not os.path.isfile(resPath):
                    mkd(os.path.dirname(resPath))
                    copyfile(hashPath, resPath)

            self.fireEvent("resource", "", count, i + 1)

    def downloadMinecraft(self):
        if not self.profile.clientDownloadUrl:
            return

        id = self.profile.id
        path = os.path.normpath(minecraft.version + "/" + id + "/" + id + ".jar")
        if not self.checkFileValidation(path, self.profile.clientHash):
            download(self.profile.clientDownloadUrl, path)

        self.fireEvent("minecraft", id, 1, 1)



import requests
import json
from pmlauncher import minecraft, mlibrary, mrule
import os


def n(t):
    if t is None:
        return ""
    else:
        return t


def arg_parse(arr):
    strlist = list()
    for item in arr:
        if type(item) == dict:
            allow = True

            rule = item.get("rules")            
            if rule:
                allow = mrule.checkAllowOS(rule)

            value = item.get("value")

            if allow and value:
                if type(value) == list:
                    strlist.extend(value)
                else:
                    strlist.append(value)
        else:
            strlist.append(item)

    return strlist



class profile:
    def __init__(self, info):
        self.id = ""
        self.assetId = ""
        self.assetUrl = ""
        self.assetHash = ""
        self.jvm_arguments = []
        self.game_arguments = []
        self.libraries = []
        self.clientDownloadUrl = ""
        self.clientHash = ""
        self.parent_profile_id = ""
        self.is_inherted = False
        self.jar = ""
        self.mainclass = ""
        self.minecraftArguments = ""
        self.releaseTime = ""
        self.type = ""
        self.parse(info)

    

    def parse(self, info):
        if info.isweb:
            json = requests.get(info.path).text
        else:
            f = open(info.path)
            json = f.read()
            f.close()

        return self.parseFromJson(json)

    def parseFromJson(self, content):
        dict = json.loads(content)

        self.id = dict.get("id")

        assetIndex = dict.get("assetIndex")
        if assetIndex:
            self.assetId = n(assetIndex.get("id"))
            self.assetUrl = n(assetIndex.get("url"))
            self.assetHash = n(assetIndex.get("sha1"))

        downloads = dict.get("downloads")
        if downloads:
            client = downloads.get("client")
            if client:
                self.clientDownloadUrl = client["url"]
                self.clientHash = client["sha1"]

        self.libraries = mlibrary.parselist(dict.get("libraries"))
        self.mainclass = n(dict.get("mainClass"))

        self.minecraftArguments = dict.get("minecraftArguments")
        arg = dict.get("arguments")
        if arg:
            if arg.get("game"):
                self.game_arguments = arg_parse(arg.get("game"))
            if arg.get("jvm"):
                self.jvm_arguments = arg_parse(arg.get("jvm"))

        self.releaseTime = n(dict.get("releaseTime"))
        self.type = n(dict.get("type"))

        inherits = dict.get("inheritsFrom")
        if inherits:
            self.is_inherited = True
            self.parent_profile_id = inherits
        else:
            self.jar = self.id

        profilePath = os.path.normpath(minecraft.version + "/" + self.id)

        if not os.path.isdir(profilePath):
            os.makedirs(profilePath)

            f = open(os.path.normpath(profilePath + "/" + self.id + ".json"), "w")
            f.write(content)
            f.close()


def inhert(parent, child):
    # Overload : assetId, assetUrl, assetHash, clientDownloadUrl, clientHash, mainClass, minecraftArguments
    # Combine : libraries, game_arguments, jvm_arguments

    if not child.assetId:
        child.assetId = parent.assetId

    if not child.assetUrl:
        child.assetUrl = parent.assetUrl

    if not child.assetHash:
        child.assetHash = parent.assetHash

    if not child.clientDownloadUrl:
        child.clientDownloadUrl = parent.clientDownloadUrl

    if not child.clientHash:
        child.clientHash = parent.clientHash

    if not child.mainClass:
        child.mainClass = parent.mainClass

    if not child.minecraftArguments:
        child.minecraftArguments = parent.minecraftArguments

    if parent.libraries:
        if child.libraries:
            child.libraries.extend(parent.libraries)
        else:
            child.libraries = parent.libraries

    if parent.game_arguments:
        if child.game_arguments:
            child.game_arguments.extend(parent.game_arguments)
        else:
            child.game_arguments = parent.game_arguments

    if parent.jvm_arguments:
        if child.jvm_arguments:
            child.jvm_arguments.extend(parent.jvm_arguments)
        else:
            child.jvm_arguments = parent.jvm_arguments


def get_profile(infos, version):
    start_profile = None

    for item in infos:
        if item.name == version:
            start_profile = profile(item)
            break
    
    if start_profile == None:
        raise ValueError("cannot find profile named " + version)

    if start_profile.is_inherted:
        parent_profile = get_profile(infos, start_profile.parent_profile_id)
        inhert(parent_profile, start_profile)

    return start_profile



# __init__.py

__all__ = [
    'mdownloader',
    'mevent',
    'minecraft',
    'mlaunchoption',
    'mlaunch',
    'mlibrary',
    'mlogin',
    'mrule',
    'mnative',
    'mprofile',
    'mprofileinfo',
    
]



from pmlauncher import minecraft
import zipfile
import os


def extract_natives(profile):
    for item in profile.libraries:
        if item.isNative:
            try:
                lib = zipfile.ZipFile(item.path)
                lib.extractall(minecraft.natives)
            except Exception as e:
                print(e)


def clean_natives():
    for item in os.listdir(minecraft.natives):
       path = os.path.join(minecraft.natives, item)
       
       if os.path.isfile(path):
            try:
                os.remove(path)
            except Exception as e:
                print(e)




from setuptools import setup, find_packages

setup(
    name="pmlauncher",
    version="0.0.3",
    url="https://github.com/AlphaBs/pml",
    license="non commercial",
    author="ksi123456ab",
    author_email="ksi123456ab@naver.com",
    description="crossplatform Python Minecraft Launcher support forge and all version",
    packages=find_packages(exclude=['main.py']),
    long_description=open('README.md').read(),
    zip_safe=False,
    install_requires=['requests'],
    python_requires='>=3',
    keywords=["minecraft", "launcher"],
    classifiers=[
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Games/Entertainment"
    ]
)



from pmlauncher import mnative, minecraft
import string
import os
import re

supportversion = "1.4"
pre_compiled = re.compile("\\$\\{(.*?)}")

def e(t):
    if " " in t:
        return '"' + t + '"'
    else:
        return t


def ea(t):
    if " " in t and "=" in t:
        s = t.split("=")
        return s[0] + '="' + s[1] + '"'
    else:
        return t


def arg_in(arg, dicts):
    args = list()
    for item in arg:
        if type(item) is str:
            m = pre_compiled.search(item)  # check ${} str
            if m:
                arg_key = m.group()  # get ${KEY}
                arg_value = dicts.get(arg_key[2:-1])  # get dicts value of ${KEY}

                if arg_value:
                    args.append(pre_compiled.sub(arg_value.replace("\\","\\\\"), item))  # replace ${} of whole str to dicts value
                else:
                    args.append(item)  # if value of default arg has space, handle whitespace.
                                           # (ex) -Dos.Version=Windows 10 => -Dos.Version="Windows 10"
            else:
                args.append(ea(item))  # not ${} str

    return args


def arg_str(arg, dicts):
    return string.Template(arg).safe_substitute(dicts)


class launch:
    def __init__(self, option):
        self.defaultJavaParameter = " ".join([
            "-XX:+UnlockExperimentalVMOptions",
            "-XX:+UseG1GC",
            "-XX:G1NewSizePercent=20",
            "-XX:G1ReservePercent=20",
            "-XX:MaxGCPauseMillis=50",
            "-XX:G1HeapRegionSize=16M"])

        option.checkValid()
        self.launchOption = option

    def createArg(self):
        profile = self.launchOption.start_profile

        args = list()

        # common jvm args
        if self.launchOption.jvm_args:
            args.append(self.launchOption.jvm_args)
        else:
            args.append(self.defaultJavaParameter)

        args.append("-Xmx" + str(self.launchOption.xmx_mb) + "m")

        # specific jvm args
        libArgs = list()

        for item in profile.libraries:
            if not item.isNative:
                libArgs.append(e(item.path))

        libArgs.append(e(os.path.normpath(minecraft.version + "/" + profile.jar + "/" + profile.jar + ".jar")))
        libs = os.pathsep.join(libArgs)

        jvmdict = {
            "natives_directory" : e(minecraft.natives),
            "launcher_name" : "minecraft-launcher",
            "launcher_version" : "2",
            "classpath" : libs
        }

        if profile.jvm_arguments:
            args.extend(arg_in(profile.jvm_arguments, jvmdict))
        else:
            args.append("-Djava.library.path=" + e(minecraft.natives))
            args.append("-cp " + libs)


        args.append(profile.mainclass)

        # game args
        gamedict = {
            "auth_player_name" : self.launchOption.session.username,
            "version_name" : profile.id,
            "game_directory" : e(minecraft.path),
            "assets_root" : e(minecraft.assets),
            "assets_index_name" : profile.assetId,
            "auth_uuid" : self.launchOption.session.uuid,
            "auth_access_token" : self.launchOption.session.access_token,
            "user_properties" : "{}",
            "user_type" : "Mojang",
            "game_assets" : e(minecraft.assetLegacy),
            "auth_session" : self.launchOption.session.access_token
        }

        if self.launchOption.launcher_name:
            gamedict["version_type"] = self.launchOption.launcher_name
        else:
            gamedict["version_type"] = profile.type

        if profile.game_arguments:  # 1.3
            args.extend(arg_in(profile.game_arguments, gamedict))
        elif profile.minecraftArguments:
            args.append(arg_str(profile.minecraftArguments, gamedict))

        # options
        if self.launchOption.server_ip:
            args.append("--server " + self.launchOption.server_ip)

        if self.launchOption.screen_width and self.launchOption.screen_height:
            args.append("--width " + self.launchOption.screen_width)
            args.append("--height ", self.launchOption.screen_height)

        return " ".join(map(str, args))

    def createProcess(self):
        mnative.clean_natives()
        mnative.extract_natives(self.launchOption.start_profile)

        return self.createArg()


